<?php

namespace BexioConnectorByHelvy;

/**
 * WooCommerce to Bexio conversion / parsing
 *
 * @since 0.1.0
 */

defined('ABSPATH') || exit;

class WCToBexio
{
    use Singleton;

    public function __construct()
    {
    }

    /**
     * Syncs a WooCommerce Order to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     * @param boolean $force_sync Whether to force the sync or not
     * @return array|WP_Error Bexio order and invoice resources
     */
    public function sync_order($order, $force_sync = false)
    {
        // Pre-sync checks
        if ($order->get_total() == 0) {
            $err_msg = "Cannot create invoice for free order (total=0)";
            return new \WP_Error('SYNC_BEXIO_ORDER', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }
        if (!get_option(AdminAPISettings::$prefix . 'sync_orders') && !$force_sync) {
            return ["status" => "SKIPPED", "reason" => "sync_orders disabled"];
        }
        if ($this->is_order_already_synced($order)) return;
        if (!$this->is_order_sync_authorized_based_on_payment_method($order) && !$force_sync) {
            return;
        }

        // Customer sync
        $contact = $this->find_or_upsert_contact_from_order($order);
        if ($contact instanceof \WP_Error) {
            return $contact;
        }

        // Customer's address(es) sync
        $billing_address = $this->upsert_contact_additional_address_from_order($contact['id'], $order, "billing");
        if ($billing_address instanceof \WP_Error) {
            return $billing_address;
        }

        $shipping_address_different_from_billing = $order->get_shipping_address_1() && ($order->get_billing_address_1() != $order->get_shipping_address_1());
        $shipping_address = $shipping_address_different_from_billing
            ? $this->upsert_contact_additional_address_from_order($contact['id'], $order, "shipping")
            : null; // = Use customers' billing address (same as shipping)
        if ($shipping_address instanceof \WP_Error) {
            return $shipping_address;
        }

        $is_tax_exempt = count($order->get_taxes()) == 0;
        $bexio_line_items = $this->woocommerce_line_items_to_bexio_positions($order, $is_tax_exempt);
        if ($bexio_line_items instanceof \WP_Error) {
            return $bexio_line_items;
        }
        $bexio_shipping_lines = $this->woocommerce_shipping_to_bexio_positions($order, $is_tax_exempt);
        if ($bexio_shipping_lines instanceof \WP_Error) {
            return $bexio_shipping_lines;
        }
        $bexio_discount_lines = $this->woocommerce_discount_to_bexio_position($order);

        // Retrieve Bexio's currency ID or create it if it doesn't exist yet in Bexio
        $bexio_currencies = API::instance()->get_currencies();
        if ($bexio_currencies instanceof \WP_Error) {
            return $bexio_currencies;
        }
        $payment_currency = strtoupper($order->get_currency());
        $currency_id = null;
        foreach ($bexio_currencies as $currency) {
            if ($payment_currency == strtoupper($currency["name"])) {
                $currency_id = $currency["id"];
                break;
            }
        }
        if (!$currency_id) {
            $bexio_currency = API::instance()->create_currency($payment_currency);
            if ($bexio_currency instanceof \WP_Error) {
                return $bexio_currency;
            }
            $currency_id = $bexio_currency["id"];
        }

        $payment_gateway = $order->get_payment_method();
        $bexio_bank_account_id = $this->get_bank_account_from_payment_gateway($payment_gateway, $payment_currency);

        // return ["bexio_bank_account" => $bexio_bank_account];
    
        // Create order in Bexio
        $bexio_order_body = [
            "title" => $this->generate_bexio_title_from_order($order),
            "is_valid_from" => $this->iso_date_to_bexio_locale_date($order->get_date_created()->__toString()),
            "contact_id" => $contact["id"],
            "contact_sub_id" => null,
            "user_id" => get_option(AdminAPISettings::$prefix . "bexio_default_user"),
            "pr_project_id" => null,
            "language_id" => get_option(AdminAPISettings::$prefix . "bexio_default_language"),
            "bank_account_id" => $bexio_bank_account_id,
            "payment_type_id" => null,
            "currency_id" => $currency_id,
            "header" => get_option(AdminAPISettings::$prefix . "invoice_header_text"),
            "footer" => get_option(AdminAPISettings::$prefix . "invoice_footer_text"),
            // Always include taxes in invoice (or exempt for shipping outside Switzerland)
            // mwst_type: isTaxExempt ? 2 : 0, // 0=including taxes; 1=excluding taxes; 2=exempt from taxes
            "mwst_type" => 0, // 0=including taxes; exemptions are handled line by line
            "mwst_is_net" => !wc_prices_include_tax(), // if type=0 -> false=Taxes are included in the total; true=Taxes will be added to the total
            "show_position_taxes" => false,
            "api_reference" => $order->get_id(),
            "template_slug" => null,
            "positions" => [
                ...$bexio_line_items,
                ...$bexio_shipping_lines,
                ...$bexio_discount_lines,
            ],
        ];

        // Handle custom - non-default - billing/shipping addresses
        if (!empty($billing_address["id"])) {
            $bexio_order_body["contact_address_id"] = $billing_address["id"];
        }
        if ($shipping_address_different_from_billing && !empty($shipping_address["id"])) {
            $bexio_order_body["delivery_address_type"] = 1; // 0 (default) = use billing address; 1 = use another address
            $bexio_order_body["delivery_address_id"] = $shipping_address["id"];
        }

        $bexio_order = API::instance()->create_order($bexio_order_body);
        if ($bexio_order instanceof \WP_Error) {
            return $bexio_order;
        }
        
        $bexio_invoice = API::instance()->create_invoice_from_order($bexio_order["id"]);
        if ($bexio_invoice instanceof \WP_Error) {
            return $bexio_invoice;
        }

        // Set invoice date to match order's date if not already the case
        if ($bexio_order["is_valid_from"] != $bexio_invoice["is_valid_from"]) {
            $updated_invoice = API::instance()->update_invoice($bexio_invoice["id"], [
                "is_valid_from" => $bexio_order_body["is_valid_from"],
            ]);
            if ($updated_invoice instanceof \WP_Error) {
                return $updated_invoice;
            }
        }

        // Directly issue invoice (i.e. don't let it as a "draft")
        $bexio_issue_invoice = API::instance()->issue_invoice($bexio_invoice["id"]);
        if ($bexio_issue_invoice instanceof \WP_Error) {
            return $bexio_issue_invoice;
        }

        // Create the delivery for this order
        $bexio_delivery = API::instance()->create_delivery_from_order($bexio_order["id"]);
        if ($bexio_delivery instanceof \WP_Error) {
            return $bexio_delivery;
        }

        // Add metafields to order
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_order_id', $bexio_order["id"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_order_number', $bexio_order["document_nr"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_invoice_id', $bexio_invoice["id"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_invoice_number', $bexio_invoice["document_nr"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_delivery_id', $bexio_delivery["id"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_delivery_number', $bexio_delivery["document_nr"]);
        $order->save_meta_data();

        return ["bexio_order" => $bexio_order, "bexio_invoice" => $bexio_invoice];
    }

    /**
     * Syncs a WooCommerce Order's payment to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     */
    public function sync_order_payment($order, $force_sync = false)
    {
        if (!$order->is_paid()) {
            $err_msg = "Order has not been paid yet";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        $bexio_invoice_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_invoice_id');
        if (!$bexio_invoice_id) {
            $err_msg = "Invoice not yet created for this order";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }
        
        // 2. Check existing payments (skip if some already exist)
        $existing_payments = API::instance()->get_invoice_payments($bexio_invoice_id);
        if ($existing_payments && is_array($existing_payments) && count($existing_payments)) {
            $err_msg = "Payment(s) already exist in Bexio for this order";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // 3. Mark invoice paid by creating the related payment
        $payment_currency = strtoupper($order->get_currency());
        $payment_gateway = $order->get_payment_method();
        $bexio_bank_account_id = $this->get_bank_account_from_payment_gateway($payment_gateway, $payment_currency);
        $payment_date = $order->get_date_paid()
            ? $this->iso_date_to_bexio_locale_date($order->get_date_paid())
            : (new \DateTime())->format('Y-m-d');
        $bexio_payment = API::instance()->create_invoice_payment($bexio_invoice_id, [
            "date" => $payment_date,
            "value" => $order->get_total(),
            "bank_account_id" => $bexio_bank_account_id,
            "payment_service_id" => null,
        ]);
        
        if ($bexio_payment instanceof \WP_Error) {
            return $bexio_payment;
        }

        // Add metafield to order
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_payment_id', $bexio_payment["id"]);
        $order->save_meta_data();
        
        return ["bexio_payment" => $bexio_payment];
    }

    /**
     * Syncs a WooCommerce Order's fulfillment to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     */
    public function sync_order_fulfillment($order, $force_sync = false)
    {
        $order_status = $order->get_status();
        if ($order_status != 'completed') {
            $err_msg = "Order has not been completed yet (status: ".$order_status.")";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        $bexio_order_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_order_id');
        if (!$bexio_order_id) {
            $err_msg = "No bexio order associated with this WooCommerce order";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }
        
        $bexio_delivery_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_delivery_id');
        if (!$bexio_delivery_id) {
            $err_msg = "No bexio delivery associated with this WooCommerce order";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }
        
        // 2. Issue the delivery (will mark the order as "Done" in bexio)
        $issued_delivery = API::instance()->issue_delivery($bexio_delivery_id);
        if ($issued_delivery instanceof \WP_Error) {
            return $issued_delivery;
        }

        // Add metafield to order
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_fulfillment_id', $bexio_delivery_id);
        $order->save_meta_data();

        return ["issued_delivery" => $issued_delivery];
    }

    /**
     * Converts an ISO 8601 date into the date format expected by bexio.
     * 
     * @param string $iso_date The ISO 8601 date to convert
     * @return string The Y-m-d formated date expected by bexio
     */
    protected function iso_date_to_bexio_locale_date($iso_date) {
        try {
            $date = new \DateTime($iso_date, new \DateTimeZone('UTC'));
            $date->setTimezone(new \DateTimeZone('Europe/Zurich'));
            return $date->format('Y-m-d');
        } catch (\Exception $e) {
            return null;
        }
    }

    /**
     * Get the bexio bank account to be used based on the given payment gateway and payment currency.
     * 
     * @param string $gateway Payment gateway method (usually a "technical" name)
     * @param string $payment_currency Currency used for the order
     * @return array ID of the bexio bank account to use
     */
    protected function get_bank_account_from_payment_gateway($gateway, $payment_currency) {
        $bexio_bank_account = get_option(AdminAPISettings::$prefix . "bexio_default_bank_account");
        $custom_payment_methods_to_bank_accounts = get_option(AdminAPISettings::$prefix . "custom_payment_methods_to_bank_accounts");
        if (!empty($custom_payment_methods_to_bank_accounts)) {
            // Check for bank account with currency matching
            $found_in_custom_bank_account_binding_with_currency = null;
            foreach ($custom_payment_methods_to_bank_accounts as $c) {
                if (strtolower($c["paymentMethod"]) === strtolower($gateway) &&
                    strtolower($c["currency"]) === strtolower($payment_currency)) {
                    $found_in_custom_bank_account_binding_with_currency = $c;
                    break;
                }
            }

            if ($found_in_custom_bank_account_binding_with_currency) {
                $bexio_bank_account = $found_in_custom_bank_account_binding_with_currency["bexioBankAccount"];
            } else {
                // Check for bank account with empty currency ("default" mapping)
                $found_in_custom_bank_account_binding = null;
                foreach ($custom_payment_methods_to_bank_accounts as $c) {
                    if (strtolower($c["paymentMethod"]) === strtolower($gateway) && empty($c["currency"])) {
                        $found_in_custom_bank_account_binding = $c;
                        break;
                    }
                }

                if ($found_in_custom_bank_account_binding) {
                    $bexio_bank_account = $found_in_custom_bank_account_binding["bexioBankAccount"];
                }
            }
        }

        return $bexio_bank_account;
    }

    /**
     * Retrieves the applicable bexio tax rate based on the given WooCommerce Order Item
     * 
     * @param object $order The WooCommerce order
     * @param object $item The order's item
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return string Applicable tax rate (bexio tax ID)
     */
    protected function get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt = false) {
        $taxes = $item->get_taxes();
        $applicable_tax = null;
        if ($is_tax_exempt || !is_array($taxes) || !count($taxes)) {
            $applicable_tax = get_option(AdminAPISettings::$prefix . 'bexio_default_exempt_tax');
        } else if (is_array($taxes) && count($taxes)) {
            // Item taxes
            $tax_rate_percentage = null;
            foreach ($taxes['total'] as $tax_rate_id => $tax_amount) {
                $tax_rate_percentage = \WC_Tax::get_rate_percent_value($tax_rate_id);
                break;
            }

            // Retrieve applicable bexio tax
            $applicable_tax = null;
            $custom_vat_mapping = get_option(AdminAPISettings::$prefix . 'custom_vat_mapping');
            if (is_array($custom_vat_mapping) && count($custom_vat_mapping)) {
                // Search applicable tax rate including country
                foreach ($custom_vat_mapping as $vat_mapping) {
                    if ($vat_mapping["vatRate"] == $tax_rate_percentage && isset($vat_mapping["country"]) &&
                        strtoupper($vat_mapping["country"]) == strtoupper($order->get_billing_country())) {
                        $applicable_tax = $vat_mapping["bexioTax"];
                        break;
                    }
                }

                if (!$applicable_tax) {
                    // Search applicable tax rate non-country specific
                    foreach ($custom_vat_mapping as $vat_mapping) {
                        if ($vat_mapping["vatRate"] == $tax_rate_percentage && empty($vat_mapping["country"])) {
                            $applicable_tax = $vat_mapping["bexioTax"];
                            break;
                        }
                    }
                }
            }

            if (!$applicable_tax) {
                return new \WP_Error("MISSING_TAX_MAPPING", "No bexio tax mapping found for " . $tax_rate_percentage . "%");
            }
        }

        return $applicable_tax;
    }


    /**
     * Converts WooCommerce's order discount to bexio line.
     * 
     * @param object $order The WooCommerce order
     * @return array Bexio positions (containing the shipping line(s))
     */
    protected function woocommerce_discount_to_bexio_position($order)
    {
        if ($order->get_discount_total()) {
            $text = get_option(AdminAPISettings::$prefix . "invoice_discount_line_text");
            $text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $text);
            $text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $text);
    
            return [
                // [
                //     "type" => "KbPositionCustom",
                //     // "amount" => 1,
                //     // "tax_id" => $applicable_tax,
                //     "text" => $text,
                //     "value" => $order->get_discount_total() * -1,
                //     "is_percentual" => false,
                //     // "discount_in_percent" => discountInPercent,
                //     // "account_id" => $account_id ? $account_id : null,
                // ]
                [
                    "type" => "KbPositionDiscount",
                    "text" => $text,
                    "value" => $order->get_discount_total(),
                    "is_percentual" => false,
                ]
            ];
        } else {
            return [];
        }
    }

    /**
     * Converts WooCommerce's shipping line to bexio positions.
     * 
     * @param object $order The WooCommerce order
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return array Bexio positions (containing the shipping line(s))
     */
    protected function woocommerce_shipping_to_bexio_positions($order, $is_tax_exempt = false)
    {
        $bexio_positions = [];
        foreach ($order->get_items('shipping') as $item) {
            $tax_included_in_prices = wc_prices_include_tax();
            $account_id = get_option(AdminAPISettings::$prefix . 'bexio_default_shipping_income_account');

            $text = get_option(AdminAPISettings::$prefix . "invoice_shipping_line_text");
            $text = str_replace('{shipping_line_title}', $item->get_name(), $text);
            $text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $text);
            $text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $text);

            $applicable_tax = $this->get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt);
            if ($applicable_tax instanceof \WP_Error) {
                return $applicable_tax;
            }
            $bexio_positions[] = [
                "type" => "KbPositionCustom",
                "amount" => 1,
                "tax_id" => $applicable_tax,
                "text" => $text,
                "unit_price" => $item->get_total(),
                // "discount_in_percent" => discountInPercent,
                "account_id" => $account_id ? $account_id : null,
            ];
        }

        return $bexio_positions;
    }

    /**
     * Converts WooCommerce's order line items to bexio positions.
     * 
     * @param object $order The WooCommerce order
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return array Bexio positions
     */
    protected function woocommerce_line_items_to_bexio_positions($order, $is_tax_exempt = false)
    {
        $bexio_positions = [];
        foreach ($order->get_items() as $item_id => $item) {

            $product = wc_get_product($item->get_product_id());
            $requires_shipping = !$product->is_virtual();

            $account_id = $requires_shipping
                ? get_option(AdminAPISettings::$prefix . 'bexio_default_product_income_account')
                : get_option(AdminAPISettings::$prefix . 'bexio_default_service_income_account');

            // Try to match item in Bexio
            $bexio_product = null;
            if ($product->get_sku()) {
                $bexio_product = API::instance()->get_article_by_sku($product->get_sku());
            }

            $text = $this->generate_line_item_text_from_product($bexio_product, $product);
            $applicable_tax = $this->get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt);
            if ($applicable_tax instanceof \WP_Error) {
                return $applicable_tax;
            }
            if ($bexio_product) {
                $bexio_position = [
                    "type" => "KbPositionArticle",
                    "article_id" => $bexio_product["id"],
                    "amount" => $item->get_quantity(),
                    "tax_id" => $applicable_tax,
                    "text" => $text,
                    "unit_price" => $product->get_price(),
                    // account_id: accountId, // -> Deprecated, will use bexio article's account_id
                ];
            } else {
                $bexio_position = [
                    "type" => "KbPositionCustom",
                    "amount" => $item->get_quantity(),
                    "tax_id" => $applicable_tax,
                    "text" => $text,
                    "unit_price" => $product->get_price(),
                    "account_id" => $account_id ? $account_id : null,
                ];
            }

            $bexio_positions[] = $bexio_position;
        }

        return $bexio_positions;
    }

    /**
     * Generates the line item text of the bexio order/invoice based on the given product
     * 
     * @param object $bexio_product Bexio product corresponding to the given product
     * @param object $wc_product Actual WooCommerce product that is part of the order
     * @return string Generated line item text
     */
    protected function generate_line_item_text_from_product($bexio_product, $wc_product)
    {
        // Log the Shopify item (this is for debugging, not usually included in production PHP code)
        $line_item_text = get_option(AdminAPISettings::$prefix . "invoice_line_item_format");
        $line_item_text = str_replace('{woocommerce_product_name}', $wc_product->get_name(), $line_item_text);
        $line_item_text = str_replace('{woocommerce_sku}', $wc_product->get_sku(), $line_item_text);
        $line_item_text = str_replace('{bexio_product_title}', !empty($bexio_product["intern_name"]) ? $bexio_product["intern_name"] : (!empty($wc_product->get_name()) ? $wc_product->get_name() : ""), $line_item_text);
        $line_item_text = str_replace('{bexio_product_description}', !empty($bexio_product["intern_description"]) ? $bexio_product["intern_description"] : "", $line_item_text);
        $line_item_text = str_replace('{bexio_sku}', !empty($bexio_product["intern_code"]) ? $bexio_product["intern_code"] : (!empty($wc_product->get_sku()) ? $wc_product->get_sku() : ""), $line_item_text);
        $line_item_text = str_replace("\n", '<br>', $line_item_text);
        $line_item_text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $line_item_text);
        $line_item_text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $line_item_text);

        return $line_item_text;
    }

    /**
     * Generates the bexio order/invoice title based on the given WooCommerce order
     * 
     * @param object $wc_order WooCommerce order
     * @return string Generated line item text
     */
    protected function generate_bexio_title_from_order($wc_order)
    {
        $bexio_title = get_option(AdminAPISettings::$prefix . "invoice_title_format");
        $bexio_title = str_replace('{woocommerce_order_id}', $wc_order->get_id(), $bexio_title);
        $bexio_title = str_replace('{woocommerce_order_number}', $wc_order->get_id(), $bexio_title);
        $bexio_title = str_replace('{woocommerce_order_title}', $wc_order->get_title(), $bexio_title);
        return $bexio_title;
    }

    /**
     * Checks if the given order is authorized to be synced or not based on the authorized payment methods
     * 
     * @param object $order The WooCommerce order
     * @return boolean
     */
    protected function is_order_sync_authorized_based_on_payment_method($order)
    {
        $payment_methods_option = get_option(AdminAPISettings::$prefix . 'sync_orders_only_for_payment_methods');
        if ($payment_methods_option) {
            $payment_methods_to_sync = array_map('trim', explode("\n", $payment_methods_option));
            $order_payment_method = $order->get_payment_method(); // Payment method slug

            // Check if the order's payment method is in the authorized list
            $found_in_payment_methods_to_sync = in_array($order_payment_method, $payment_methods_to_sync);
            if (!$found_in_payment_methods_to_sync) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks if the given order has already been synced
     * 
     * @param object $order The WooCommerce order
     * @return boolean
     */
    protected function is_order_already_synced($order)
    {
        return $order->get_meta(AdminAPISettings::$prefix . 'bexio_invoice_id')
            || $order->get_meta(AdminAPISettings::$prefix . 'bexio_order_id');
    }

    /**
     * Retrieves the bexio contact (find / upsert) corresponding to the given order's customer
     * 
     * @param object $order The WooCommerce order
     * @return array|WP_Error Bexio contact corresponding to the order's customer
     */
    protected function find_or_upsert_contact_from_order($order)
    {
        $company_name = $order->get_billing_company();

        // Fetch matching bexio contact(s) by email, otherwise by first and last name
        $customer_email = $order->get_billing_email();
        $customer_first_name = $order->get_billing_first_name();
        $customer_last_name = $order->get_billing_last_name();
        $contacts = API::instance()->get_contacts_by_email_or_name($customer_email, $customer_first_name, $customer_last_name);
        if ($contacts instanceof \WP_Error) {
            return $contacts;
        }

        $person_contact = null;
        $company_contact = null; // Both contacts to fill
        if (count($contacts) > 0) {
            foreach ($contacts as $contact) {
                if ($contact['contact_type_id'] == 2) {
                    $person_contact = $contact;
                }
            }
            foreach ($contacts as $contact) {
                if ($contact['contact_type_id'] == 1) {
                    $company_contact = $contact;
                }
            }
        }

        // If the person contact is found, update it; otherwise, create it
        $person_contact = $this->upsert_customer($order, $person_contact);
        if ($person_contact instanceof \WP_Error) {
            return $person_contact;
        }

        // Handle company contact / information
        if ($company_name) {
            // If available, update the company information and link the company to the order, not the physical person
            if (!$company_contact) {
                // Search contacts owning this person (i.e. their company)
                $company_relations = API::instance()->search_contact_relations([["field" => "contact_sub_id", "value" => $person_contact['id'], "criteria" => "="]]);
                if ($company_relations instanceof \WP_Error) {
                    return $company_relations;
                }
                if (is_array($company_relations) && count($company_relations)) {
                    foreach ($company_relations as $company_relation) {
                        $contact = API::instance()->get_contact_by_id($company_relation['contact_id']);
                        if ($contact instanceof \WP_Error) {
                            return $contact;
                        }
                        if (strtolower($contact['name_1']) == strtolower($company_name)) {
                            $company_contact = $contact;
                            break;
                        }
                    }
                }

                // Otherwise, search companyContact directly by name & city (not name only to prevent match on homonyms)
                if (!$company_contact) {
                    $company_contacts = API::instance()->get_contacts_companies_by_name_and_postcode($company_name, $order->get_billing_postcode());
                    if ($company_contacts instanceof \WP_Error) {
                        return $company_contacts;
                    }
                    if (count($company_contacts)) {
                        $company_contact = $company_contacts[0];
                        // In this case, create the relation between the person and the company in bexio
                        $contact_relation = API::instance()->create_contact_relation($company_contact['id'], $person_contact['id']);
                        if ($contact_relation instanceof \WP_Error) {
                            return $contact_relation;
                        }
                    }
                }
            }

            // If found, update company's contact information; otherwise, create the company contact and the relation with the existing customer
            $company_contact = $this->upsert_customer($order, $company_contact, $person_contact['id']);
            if ($company_contact instanceof \WP_Error) {
                return $company_contact;
            }
        }

        return $company_contact
            ? $company_contact // The customer is a company
            : $person_contact; // The customer is a private person
    }

    /**
     * Upsert the customer of the given WooCommerce order.
     * If the bexio contact is empty, the contact is created in bexio.
     * Otherwise, the given bexio contact is updated based on the order's customer information.
     * 
     * @param object $order The WooCommerce order
     * @param object $bexio_contact The contact corresponding to the order's customer, if already found in bexio contacts
     * @param string $is_company_owning_contact_id Provide a value if the given contact is a company that owns the given contact ID (for binding)
     * @return array Upserted bexio contact
     */
    protected function upsert_customer($order, $bexio_contact, $is_company_owning_contact_id = null)
    {
        error_log("Upserting customer with bexio_default_contacts_group : " . get_option(AdminAPISettings::$prefix . 'bexio_default_contacts_group'));
        $company_name = $order->get_billing_company();

        $billing_address_line = $order->get_billing_address_1();
        if ($order->get_billing_address_2()) {
            $billing_address_line .= "\n" . $order->get_billing_address_2();
        }

        if ($bexio_contact) {
            $contact_data = []; // Potential customer's data to update

            // Update contact's default (billing) address if it has changed
            if (!$bexio_contact['address'] || ($billing_address_line && $bexio_contact['address'] != $billing_address_line)) {
                $billing_address_country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
                if ($billing_address_country instanceof \WP_Error) {
                    return $billing_address_country;
                }

                $contact_data["address"] = $billing_address_line;
                $contact_data["postcode"] = $order->get_billing_postcode();
                $contact_data["city"] = $order->get_billing_city();
                $contact_data["country_id"] = $billing_address_country['id'];
            }

            // Update customer's mobile phone if needed
            if ($order->get_billing_phone() && $bexio_contact["phone_mobile"] != $order->get_billing_phone()) {
                $contact_data["phone_mobile"] = $order->get_billing_phone();
            }

            // Update Bexio customer (if some data has changed)
            if (count($contact_data) > 0) {
                $bexio_contact = API::instance()->update_contact_by_id($bexio_contact['id'], $bexio_contact);
                if ($bexio_contact instanceof \WP_Error) {
                    return $bexio_contact;
                }
            }
        } else {
            $billing_address_country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
            if ($billing_address_country instanceof \WP_Error) {
                return $billing_address_country;
            }

            if ($is_company_owning_contact_id) {
                // Create a company
                $contact_object = [
                    "contact_type_id" => 1,
                    "name_1" => $company_name,
                    "address" => $billing_address_line,
                    "postcode" => $order->get_billing_postcode(),
                    "city" => $order->get_billing_city(),
                    "country_id" => $billing_address_country['id'],
                    "contact_group_ids" => get_option(AdminAPISettings::$prefix . 'bexio_default_contacts_group'), // If list, coma-separated string, e.g. "1,2"
                    "user_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                    "owner_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                ];
                $bexio_contact = API::instance()->create_contact($contact_object);
                if (!empty($bexio_contact['id'])) {
                    API::instance()->create_contact_relation($bexio_contact['id'], $is_company_owning_contact_id);
                }
            } else {
                // Create a physical person
                $bexio_contact = API::instance()->create_contact([
                    "contact_type_id" => 2,
                    "name_1" => $order->get_billing_last_name(), // Empty string if no name provided (prevent bexio error)
                    "name_2" => $order->get_billing_first_name(),
                    "salutation_id" => null,
                    "salutation_form" => null,
                    "titel_id" => null,
                    "birthday" => null,
                    "address" => $billing_address_line,
                    "postcode" => $order->get_billing_postcode(),
                    "city" => $order->get_billing_city(),
                    "country_id" => $billing_address_country['id'],
                    "mail" => $order->get_billing_email(),
                    "phone_fixed" => null,
                    "phone_fixed_second" => null,
                    "phone_mobile" => $order->get_billing_phone(),
                    "fax" => null,
                    "url" => null,
                    "skype_name" => null,
                    "remarks" => null,
                    "language_id" => null,
                    "contact_group_ids" => get_option(AdminAPISettings::$prefix . 'bexio_default_contacts_group'), // If list, coma-separated string, e.g. "1,2"
                    "contact_branch_ids" => null,
                    "user_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                    "owner_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                ]);
            }
        }

        return $bexio_contact;
    }

    /**
     * Converts an order's address (billing or shipping) to bexio address format
     * 
     * @param object $order The WooCommerce order resource
     * @param string $address_type The type of address to convert (billing or shipping, defaults to billing)
     * @return array Bexio-structured address
     */
    protected function order_address_to_bexio_address($order, $address_type = "billing")
    {
        if ($address_type == "shipping") {
            $address = $order->get_shipping_address_1();
            if ($order->get_shipping_address_2()) {
                $address .= "\n" . $order->get_shipping_address_2();
            }

            $country = API::instance()->find_or_create_country_by_country_code($order->get_shipping_country());
            $name = $order->get_shipping_company()
                ? $order->get_shipping_company() . "\n" . $order->get_formatted_shipping_full_name()
                : $order->get_formatted_shipping_full_name();

            return [
                "name" => $name,
                "address" => $address,
                "postcode" => $order->get_shipping_postcode(),
                "city" => $order->get_shipping_city(),
                "country_id" => $country['id'],
            ];
        } else { // Billing
            $address = $order->get_billing_address_1();
            if ($order->get_billing_address_2()) {
                $address .= "\n" . $order->get_billing_address_2();
            }
            $country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
            $name = $order->get_billing_company()
                ? $order->get_billing_company() . "\n" . $order->get_formatted_billing_full_name()
                : $order->get_formatted_billing_full_name();

            return [
                "name" => $name,
                "address" => $address,
                "postcode" => $order->get_billing_postcode(),
                "city" => $order->get_billing_city(),
                "country_id" => $country['id'],
            ];
        }
    }

    /**
     * Updates the bexio contact with the given order's address
     * 
     * @param string $contact_id Bexio contact ID
     * @param object $order The WooCommerce order resource
     * @param string $address_type Whether the address is "billing" or "shipping"
     * @return array Created or updated additional adress
     */
    protected function upsert_contact_additional_address_from_order($contact_id, $order, $address_type)
    {
        $address_subject = "WooCommerce " . $address_type . " address";
        $additional_addresses = API::instance()->get_contact_additional_addresses_by_subject($contact_id, $address_subject);
        $existing_address_id = count($additional_addresses) ? $additional_addresses[0]['id'] : 0;

        $bexio_address_obj = $this->order_address_to_bexio_address($order, $address_type);
        $bexio_address_obj['subject'] = $address_subject;
        $bexio_address_obj['description'] = $address_subject . " of the customer - Last update " . (new \DateTime())->format('c');

        return $existing_address_id
            ? API::instance()->update_contact_additional_address($contact_id, $existing_address_id, $bexio_address_obj)
            : API::instance()->create_contact_additional_addresses($contact_id, $bexio_address_obj);
    }

    /**
     * Cloning is forbidden.
     */
    public function __clone()
    {
        _doing_it_wrong(__FUNCTION__, __('Cheatin&#8217; huh?'), BEXIO_CONNECTOR_BY_HELVY_VERSION);
    }
}
